import math
import torch
import torch.nn.functional as F
from torch import nn
from einops import rearrange
from mambax.pscan import pscan


import math
import torch
import torch.nn as nn
import torch.nn.functional as F


class Mamba(nn.Module):
    def __init__(
        self,
        d_model,
        d_inner,
        d_conv=4,
        d_state=16,
        dt_rank=64,
        act=nn.SiLU()
    ):
        super().__init__()
        self.d_model = d_model
        self.d_inner = d_inner
        self.d_conv = d_conv
        self.d_state = d_state
        self.dt_rank = dt_rank
        self.act = act

        # Projections
        self.in_proj = nn.Linear(d_model, 2 * d_inner, bias=False)
        self.conv1d = nn.Conv1d(
            d_inner,
            d_inner,
            kernel_size=d_conv,
            bias=False,
            groups=d_inner,
            padding=d_conv - 1
        )
        self.x_proj = nn.Linear(d_inner, dt_rank + 2 * d_state, bias=False)
        self.dt_proj = nn.Linear(dt_rank, d_inner, bias=True)
        
        # Initialize dt_proj
        dt_std = dt_rank ** -0.5
        nn.init.uniform_(self.dt_proj.weight, -dt_std, dt_std)

        # Initialize dt bias
        dt = torch.exp(
            torch.rand(d_inner) * (math.log(0.1) - math.log(1e-3)) + math.log(1e-3)
        )
        inv_dt = dt + torch.log(-torch.expm1(-dt))
        with torch.no_grad():
            self.dt_proj.bias.copy_(inv_dt)

        # State parameters
        A = torch.arange(1, d_state + 1).repeat(d_inner, 1).float()
        self.A_log = nn.Parameter(torch.log(A))
        self.A_log._no_weight_decay = True

        self.D = nn.Parameter(torch.ones(d_inner))
        self.D._no_weight_decay = True

        # Output projections
        self.out_proj = nn.Linear(d_inner, d_model, bias=False)
        self.drop = nn.Dropout(0.1)

    def forward(self, x, h0=None, conv_cache=None, onnx_export=True):
        """Forward pass of Mamba block.
        
        Args:
            x: Input tensor of shape (B, L, D)
            h0: Initial hidden state of shape (B, ED, N)
            conv_cache: Convolution cache of shape (B, ED, d_conv-1)
            onnx_export: Whether to export to ONNX format
        """
        if h0 is None:
            return self._forward_full(x)
        else:
            return self._forward_cumprod(x, h0, conv_cache, onnx_export)

    def _forward_cumprod(self, x, h0=None, conv_cache=None, onnx_export=True):
      
        xz = self.in_proj(x)  # (B, Lc, 2*ED)
        x_part, z_part = xz.chunk(2, dim=-1)  # (B, Lc, ED)

        x_cat = torch.cat([conv_cache, x_part.transpose(1, 2)], dim=2)  # (B, ED, d_conv-1+Lc)
        x_conv = F.conv1d(
            x_cat, self.conv1d.weight, groups=self.d_inner
        )  # (B, ED, Lc)
        x_conv = self.act(x_conv.transpose(1, 2))  # (B, Lc, ED)

        A = -torch.exp(self.A_log.float())
        D = self.D
        deltaBC = self.x_proj(x_conv)  # (B, Lc, dt_rank+2N)
        delta, B_, C_ = torch.split(
            deltaBC,
            [self.dt_rank, self.d_state, self.d_state],
            dim=-1
        )
        
        delta = self.dt_proj.weight @ delta.transpose(1, 2)  # (B, Lc, ED)
        delta = delta.transpose(1, 2)
        delta = F.softplus(delta + self.dt_proj.bias)
      
        deltaA = torch.exp(delta.unsqueeze(-1) * A)  # (B, Lc, ED, N)
        deltaB = delta.unsqueeze(-1) * B_.unsqueeze(2)  # (B, Lc, ED, N)
        BX = deltaB * x_conv.unsqueeze(-1)  # (B, Lc, ED, N)

        if onnx_export:
            P = torch.exp(torch.cumsum(torch.log(deltaA + 1e-8), dim=1))
        else:
            P = torch.cumprod(deltaA, dim=1) + 1e-8

        invP = 1.0 / P
        S = torch.cumsum(BX * invP, dim=1)  
        h_all = P * (h0.unsqueeze(1) + S)  # (B, Lc, ED, N)
        h_new = h_all[:, -1]  # (B, ED, N) — в cache

        y = (h_all @ C_.unsqueeze(-1)).squeeze(-1) + D * x_conv  # (B, Lc, ED)
        z_part = self.act(z_part)
        out = self.out_proj(y * z_part)  # (B, Lc, D)

        new_inputs = torch.cat([conv_cache, x_part.transpose(1, 2)], dim=2)
        new_inputs = new_inputs[:, :, -(self.d_conv - 1):]

        return out, h_new, new_inputs

    def _forward_full(self, x):
        _, L, _ = x.shape
        
        xz = self.in_proj(x)
        x, z = xz.chunk(2, dim=-1)

        x = x.transpose(1, 2)
        x = self.conv1d(x)[:, :, :L]
        x = x.transpose(1, 2)
        x = self.act(x)
        
        y = self.ssm(x)
        z = self.act(z)

        out = self.out_proj(y * z)
        if self.training:
            out = self.drop(out)
        return out

    def ssm(self, x):
        A = -torch.exp(self.A_log.float())
        D = self.D
        
        deltaBC = self.x_proj(x)
        delta, B, C = torch.split(
            deltaBC,
            [self.dt_rank, self.d_state, self.d_state],
            dim=-1
        )
        
        delta = self.dt_proj.weight @ delta.transpose(1, 2)
        delta = delta.transpose(1, 2)
        delta = F.softplus(delta + self.dt_proj.bias)
        
        y = self.selective_scan(x, delta, A, B, C, D)
        return y

    def selective_scan(self, x, delta, A, B, C, D):
        deltaA = torch.exp(delta.unsqueeze(-1) * A)  # (B, L, ED, N)
        deltaB = delta.unsqueeze(-1) * B.unsqueeze(2)  # (B, L, ED, N)
        BX = deltaB * x.unsqueeze(-1)

        hs = pscan(deltaA, BX)
        y = (hs @ C.unsqueeze(-1)).squeeze(3)
        return y + D * x



import math
import torch
import torch.nn.functional as F
from torch import nn


def npo2(L):                       
    """
    Returns the next power of 2 above len
    """
    return 1 << math.ceil(math.log2(L))

def pad_npo2(X):
    """
    Pads input length dim to the next power of 2

    Args:
        X : (B, L, D, N)

    Returns:
        Y : (B, npo2(L), D, N)
    """
    Lp = npo2(X.size(1))
    pad = (0, 0, 0, 0, 0, Lp - X.size(1))
    return F.pad(X, pad, value=0.)


class PScan(torch.autograd.Function):
    @staticmethod
    def pscan(A, X):
        # A : (B, D, L, N)
        # X : (B, D, L, N)

        # modifies X in place by doing a parallel scan.
        # more formally, X will be populated by these values :
        # H[t] = A[t] * H[t-1] + X[t] with H[0] = 0
        # which are computed in parallel (2*log2(T) sequential steps (ideally), instead of T sequential steps)

        # only supports L that is a power of two (mainly for a clearer code)
        
        B, D, L, _ = A.size()
        num_steps = int(math.log2(L))

        # up sweep (last 2 steps unfolded)
        Aa = A
        Xa = X
        for _ in range(num_steps-2):
            T = Xa.size(2)
            Aa = Aa.view(B, D, T//2, 2, -1)
            Xa = Xa.view(B, D, T//2, 2, -1)
            
            Xa[:, :, :, 1].add_(Aa[:, :, :, 1].mul(Xa[:, :, :, 0]))
            Aa[:, :, :, 1].mul_(Aa[:, :, :, 0])

            Aa = Aa[:, :, :, 1]
            Xa = Xa[:, :, :, 1]

        # we have only 4, 2 or 1 nodes left
        if Xa.size(2) == 4:
            Xa[:, :, 1].add_(Aa[:, :, 1].mul(Xa[:, :, 0]))
            Aa[:, :, 1].mul_(Aa[:, :, 0])

            Xa[:, :, 3].add_(Aa[:, :, 3].mul(Xa[:, :, 2] + Aa[:, :, 2].mul(Xa[:, :, 1])))
        elif Xa.size(2) == 2:
            Xa[:, :, 1].add_(Aa[:, :, 1].mul(Xa[:, :, 0]))
            return
        else:
            return

        # down sweep (first 2 steps unfolded)
        Aa = A[:, :, 2**(num_steps-2)-1:L:2**(num_steps-2)]
        Xa = X[:, :, 2**(num_steps-2)-1:L:2**(num_steps-2)]
        Xa[:, :, 2].add_(Aa[:, :, 2].mul(Xa[:, :, 1]))
        Aa[:, :, 2].mul_(Aa[:, :, 1])

        for k in range(num_steps-3, -1, -1):
            Aa = A[:, :, 2**k-1:L:2**k]
            Xa = X[:, :, 2**k-1:L:2**k]

            T = Xa.size(2)
            Aa = Aa.view(B, D, T//2, 2, -1)
            Xa = Xa.view(B, D, T//2, 2, -1)

            Xa[:, :, 1:, 0].add_(Aa[:, :, 1:, 0].mul(Xa[:, :, :-1, 1]))
            Aa[:, :, 1:, 0].mul_(Aa[:, :, :-1, 1])

    @staticmethod
    def pscan_rev(A, X):
        # A : (B, D, L, N)
        # X : (B, D, L, N)

        # the same function as above, but in reverse
        # (if you flip the input, call pscan, then flip the output, you get what this function outputs)
        # it is used in the backward pass

        # only supports L that is a power of two (mainly for a clearer code)

        B, D, L, _ = A.size()
        num_steps = int(math.log2(L))

        # up sweep (last 2 steps unfolded)
        Aa = A
        Xa = X
        for _ in range(num_steps-2):
            T = Xa.size(2)
            Aa = Aa.view(B, D, T//2, 2, -1)
            Xa = Xa.view(B, D, T//2, 2, -1)
                    
            Xa[:, :, :, 0].add_(Aa[:, :, :, 0].mul(Xa[:, :, :, 1]))
            Aa[:, :, :, 0].mul_(Aa[:, :, :, 1])

            Aa = Aa[:, :, :, 0]
            Xa = Xa[:, :, :, 0]

        # we have only 4, 2 or 1 nodes left
        if Xa.size(2) == 4:
            Xa[:, :, 2].add_(Aa[:, :, 2].mul(Xa[:, :, 3]))
            Aa[:, :, 2].mul_(Aa[:, :, 3])

            Xa[:, :, 0].add_(Aa[:, :, 0].mul(Xa[:, :, 1].add(Aa[:, :, 1].mul(Xa[:, :, 2]))))
        elif Xa.size(2) == 2:
            Xa[:, :, 0].add_(Aa[:, :, 0].mul(Xa[:, :, 1]))
            return
        else:
            return

        # down sweep (first 2 steps unfolded)
        Aa = A[:, :, 0:L:2**(num_steps-2)]
        Xa = X[:, :, 0:L:2**(num_steps-2)]
        Xa[:, :, 1].add_(Aa[:, :, 1].mul(Xa[:, :, 2]))
        Aa[:, :, 1].mul_(Aa[:, :, 2])

        for k in range(num_steps-3, -1, -1):
            Aa = A[:, :, 0:L:2**k]
            Xa = X[:, :, 0:L:2**k]

            T = Xa.size(2)
            Aa = Aa.view(B, D, T//2, 2, -1)
            Xa = Xa.view(B, D, T//2, 2, -1)

            Xa[:, :, :-1, 1].add_(Aa[:, :, :-1, 1].mul(Xa[:, :, 1:, 0]))
            Aa[:, :, :-1, 1].mul_(Aa[:, :, 1:, 0])

    @staticmethod
    def forward(ctx, A_in, X_in):
        """
        Applies the parallel scan operation, as defined above. Returns a new tensor.
        If you can, privilege sequence lengths that are powers of two.

        Args:
            A_in : (B, L, D, N)
            X_in : (B, L, D, N)

        Returns:
            H : (B, L, D, N)
        """

        L = X_in.size(1)

        # cloning is requiered because of the in-place ops
        if L == npo2(L):
            A = A_in.clone()
            X = X_in.clone()
        else:
            # pad tensors (and clone btw)
            A = pad_npo2(A_in) # (B, npo2(L), D, N)
            X = pad_npo2(X_in) # (B, npo2(L), D, N)
        
        # prepare tensors
        A = A.transpose(2, 1) # (B, D, npo2(L), N)
        X = X.transpose(2, 1) # (B, D, npo2(L), N)

        # parallel scan (modifies X in-place)
        PScan.pscan(A, X)

        ctx.save_for_backward(A_in, X)
        
        # slice [:, :L] (cut if there was padding)
        return X.transpose(2, 1)[:, :L]
    
    @staticmethod
    def backward(ctx, grad_output_in):
        """
        Flows the gradient from the output to the input. Returns two new tensors.

        Args:
            ctx : A_in : (B, L, D, N), X : (B, D, L, N)
            grad_output_in : (B, L, D, N)

        Returns:
            gradA : (B, L, D, N), gradX : (B, L, D, N)
        """

        A_in, X = ctx.saved_tensors

        L = grad_output_in.size(1)

        # cloning is requiered because of the in-place ops
        if L == npo2(L):
            grad_output = grad_output_in.clone()
            # the next padding will clone A_in
        else:
            grad_output = pad_npo2(grad_output_in) # (B, npo2(L), D, N)
            A_in = pad_npo2(A_in) # (B, npo2(L), D, N)

        # prepare tensors
        grad_output = grad_output.transpose(2, 1)
        A_in = A_in.transpose(2, 1) # (B, D, npo2(L), N)
        A = F.pad(A_in[:, :, 1:], (0, 0, 0, 1)) # (B, D, npo2(L), N) shift 1 to the left (see hand derivation)

        # reverse parallel scan (modifies grad_output in-place)
        PScan.pscan_rev(A, grad_output)

        Q = torch.zeros_like(X)
        Q[:, :, 1:].add_(X[:, :, :-1] * grad_output[:, :, 1:])

        return Q.transpose(2, 1)[:, :L], grad_output.transpose(2, 1)[:, :L]
    
pscan = PScan.apply



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="mambax",
    version="0.1.0",
    packages=find_packages(),
    install_requires=["torch>=2.0.0", "numpy>=1.21.0"],
    author="Oleg Kufa",
    author_email="os.schischkin@gmail.com",
    description="Optimized Mamba implementation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jootanehorror/MambaX",
    license="MIT",  
    python_requires=">=3.10",
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
    ],
)
